@@ -21,9 +21,9 @@
 # THE SOFTWARE.
 
 
-__version__ = '0.9.6.2'
+__version__ = '0.9.7.1'
 
 
 def embed(locals_=None, args=['-i', '-q'], banner=None):
     from bpython.cli import main
-    return main(args, locals_)
+    return main(args, locals_, banner)
