@@ -7,12 +7,21 @@ import os
 import sys
 import code
 from optparse import OptionParser, OptionGroup
-from itertools import takewhile
 
 from bpython import __version__
 from bpython.config import loadini, Struct, migrate_rc
 
-def parse(args, extras=None):
+
+class OptionParserFailed(ValueError):
+    """Raised by the RaisingOptionParser for a bogus commandline."""
+
+
+class RaisingOptionParser(OptionParser):
+    def error(self, msg):
+        raise OptionParserFailed()
+
+
+def parse(args, extras=None, ignore_stdin=False):
     """Receive an argument list - if None, use sys.argv - parse all args and
     take appropriate action. Also receive optional extra options: this should
     be a tuple of (title, description, options)
@@ -39,10 +48,16 @@ def parse(args, extras=None):
     if args is None:
         args = sys.argv[1:]
 
-    parser = OptionParser(usage='Usage: %prog [options] [file [args]]\n'
-                                'NOTE: If bpython sees an argument it does '
-                                 'not know, execution falls back to the '
-                                 'regular Python interpreter.')
+    parser = RaisingOptionParser(
+        usage='Usage: %prog [options] [file [args]]\n'
+        'NOTE: If bpython sees an argument it does '
+        'not know, execution falls back to the '
+        'regular Python interpreter.')
+    # This is not sufficient if bpython gains its own -m support
+    # (instead of falling back to Python itself for that).
+    # That's probably fixable though, for example by having that
+    # option swallow all remaining arguments in a callback.
+    parser.disable_interspersed_args()
     parser.add_option('--config', '-c', default='~/.bpython/config',
                       help='use CONFIG instead of default config file')
     parser.add_option('--interactive', '-i', action='store_true',
@@ -59,26 +74,20 @@ def parse(args, extras=None):
             extras_group.add_option(option)
         parser.add_option_group(extras_group)
 
-    all_args = set(parser._short_opt.keys() + parser._long_opt.keys())
-    if args and not all_args.intersection(arg.split('=')[0] for arg in args):
+    try:
+        options, args = parser.parse_args(args)
+    except OptionParserFailed:
         # Just let Python handle this
         os.execv(sys.executable, [sys.executable] + args)
-    else:
-        # Split args in bpython args and args for the executed file
-        real_args = list(takewhile(lambda arg: arg.split('=')[0] in all_args,
-                                   args))
-        exec_args = args[len(real_args):]
-
-    options, args = parser.parse_args(real_args)
 
     if options.version:
         print 'bpython version', __version__,
         print 'on top of Python', sys.version.split()[0]
-        print ('(C) 2008-2009 Bob Farrell, Andreas Stuehrk et al. '
+        print ('(C) 2008-2010 Bob Farrell, Andreas Stuehrk et al. '
                'See AUTHORS for detail.')
         raise SystemExit
 
-    if not (sys.stdin.isatty() and sys.stdout.isatty()):
+    if not ignore_stdin and not (sys.stdin.isatty() and sys.stdout.isatty()):
         interpreter = code.InteractiveInterpreter()
         interpreter.runsource(sys.stdin.read())
         raise SystemExit
@@ -91,7 +100,7 @@ def parse(args, extras=None):
 
     loadini(config, options.config)
 
-    return config, options, exec_args
+    return config, options, args
 
 def exec_code(interpreter, args):
     """
@@ -101,5 +110,6 @@ def exec_code(interpreter, args):
     with open(args[0], 'r') as sourcefile:
         code_obj = compile(sourcefile.read(), args[0], 'exec')
     old_argv, sys.argv = sys.argv, args
+    sys.path.insert(0, os.path.abspath(os.path.dirname(args[0])))
     interpreter.runcode(code_obj)
     sys.argv = old_argv
