@@ -1,5 +1,4 @@
-#!/usr/bin/env python
-#
+
 # The MIT License
 #
 # Copyright (c) 2008 Bob Farrell
@@ -31,7 +30,6 @@ import curses
 import math
 import re
 import time
-import inspect
 import signal
 import struct
 import termios
@@ -39,7 +37,8 @@ import fcntl
 import unicodedata
 import errno
 
-from locale import LC_ALL, getpreferredencoding, setlocale
+from locale import LC_ALL, setlocale
+import locale
 from types import ModuleType
 
 # These are used for syntax hilighting.
@@ -50,7 +49,6 @@ from pygments.token import Token
 from bpython.formatter import BPythonFormatter
 
 # This for completion
-from bpython import inspection
 from bpython import importcompletion
 
 # This for config
@@ -59,11 +57,10 @@ from bpython.config import Struct
 # This for keys
 from bpython.keys import key_dispatch
 
+from bpython import repl
 from bpython.pager import page
-from bpython.repl import Interpreter, Repl
 import bpython.args
 
-
 def log(x):
     f = open('/tmp/bpython.log', 'a')
     f.write('%s\n' % (x,))
@@ -71,6 +68,8 @@ def log(x):
 py3 = sys.version_info[0] == 3
 stdscr = None
 
+def getpreferredencoding():
+    return locale.getpreferredencoding() or sys.getdefaultencoding()
 
 def calculate_screen_lines(tokens, width, cursor=0):
     """Given a stream of tokens and a screen width plus an optional
@@ -234,12 +233,34 @@ def make_colors(config):
     return c
 
 
-class CLIRepl(Repl):
+class CLIInteraction(repl.Interaction):
+    def __init__(self, config, statusbar=None):
+        repl.Interaction.__init__(self, config, statusbar)
+
+    def confirm(self, q):
+        """Ask for yes or no and return boolean"""
+        try:
+            reply = self.statusbar.prompt(q)
+        except ValueError:
+            return False
+
+        return reply.lower() in ('y', 'yes')
+
+
+    def notify(self, s, n=10):
+        return self.statusbar.message(s, n)
+
+    def file_prompt(self, s):
+        return self.statusbar.prompt(s)
+
+
+class CLIRepl(repl.Repl):
 
     def __init__(self, scr, interp, statusbar, config, idle=None):
-        Repl.__init__(self, interp, config, idle)
-        interp.writetb = self.writetb
+        repl.Repl.__init__(self, interp, config)
+        self.interp.writetb = self.writetb
         self.scr = scr
+        self.stdout_hist = ''
         self.list_win = newwin(get_colpair(config, 'background'), 1, 1, 1, 1)
         self.cpos = 0
         self.do_exit = False
@@ -250,6 +271,8 @@ class CLIRepl(Repl):
         self.last_key_press = time.time()
         self.s = ''
         self.statusbar = statusbar
+        self.formatter = BPythonFormatter(config.color_scheme)
+        self.interact = CLIInteraction(self.config, statusbar=self.statusbar)
 
     def addstr(self, s):
         """Add a string to the current input line and figure out
@@ -271,16 +294,6 @@ class CLIRepl(Repl):
 
         return not self.s.lstrip()
 
-    def back(self):
-        """Replace the active line with previous line in history and
-        increment the index to keep track"""
-
-        self.cpos = 0
-        self.rl_history.enter(self.s)
-        self.clear_wrapped_lines()
-        self.s = self.rl_history.back()
-        self.print_line(self.s, clr=True)
-
     def bs(self, delete_tabs=True):
         """Process a backspace"""
 
@@ -313,13 +326,18 @@ class CLIRepl(Repl):
 
     def bs_word(self):
         pos = len(self.s) - self.cpos - 1
+        deleted = []
 # First we delete any space to the left of the cursor.
         while pos >= 0 and self.s[pos] == ' ':
+            deleted.append(self.s[pos])
             pos -= self.bs()
 # Then we delete a full word.
         while pos >= 0 and self.s[pos] != ' ':
+            deleted.append(self.s[pos])
             pos -= self.bs()
 
+        return ''.join(reversed(deleted))
+
     def check(self):
         """Check if paste mode should still be active and, if not, deactivate
         it and force syntax highlighting."""
@@ -333,14 +351,15 @@ class CLIRepl(Repl):
         """Called when a SyntaxError occured in the interpreter. It is
         used to prevent autoindentation from occuring after a
         traceback."""
-        Repl.clear_current_line(self)
+        repl.Repl.clear_current_line(self)
         self.s = ''
 
     def clear_wrapped_lines(self):
         """Clear the wrapped lines of the current input."""
         # curses does not handle this on its own. Sad.
-        width = self.scr.getmaxyx()[1]
-        for y in xrange(self.iy + 1, self.iy + len(self.s) // width + 1):
+        height, width = self.scr.getmaxyx()
+        max_y = min(self.iy + (self.ix + len(self.s)) // width + 1, height)
+        for y in xrange(self.iy + 1, max_y):
             self.scr.move(y, 0)
             self.scr.clrtoeol()
 
@@ -357,7 +376,7 @@ class CLIRepl(Repl):
             return
 
         if self.config.auto_display_list or tab:
-            self.list_win_visible = Repl.complete(self, tab)
+            self.list_win_visible = repl.Repl.complete(self, tab)
             if self.list_win_visible:
                 try:
                     self.show_list(self.matches, self.argspec)
@@ -455,8 +474,11 @@ class CLIRepl(Repl):
                     a = a | curses.A_BOLD
         s = s.replace('\x03', '')
         s = s.replace('\x01', '')
+
         # Replace NUL bytes, as addstr raises an exception otherwise
-        s = s.replace('\x00', '')
+        s = s.replace('\0', '')
+        # Replace \r\n bytes, as addstr remove the current line otherwise
+        s = s.replace('\r\n', '\n')
 
         self.scr.addstr(s, a)
 
@@ -474,16 +496,38 @@ class CLIRepl(Repl):
 
         return True
 
+    def hbegin(self):
+        """Replace the active line with first line in history and
+        increment the index to keep track"""
+        self.cpos = 0
+        self.clear_wrapped_lines()
+        self.rl_history.enter(self.s)
+        self.s = self.rl_history.first()
+        self.print_line(self.s, clr=True)
+
+    def hend(self):
+        """Same as hbegin() but, well, forward"""
+        self.cpos = 0
+        self.clear_wrapped_lines()
+        self.rl_history.enter(self.s)
+        self.s = self.rl_history.last()
+        self.print_line(self.s, clr=True)
+
+    def back(self):
+        """Replace the active line with previous line in history and
+        increment the index to keep track"""
+
+        self.cpos = 0
+        self.clear_wrapped_lines()
+        self.rl_history.enter(self.s)
+        self.s = self.rl_history.back()
+        self.print_line(self.s, clr=True)
+
     def fwd(self):
         """Same as back() but, well, forward"""
 
         self.cpos = 0
-
-        width = self.scr.getmaxyx()[1]
-        for y in xrange(self.iy + 1, self.iy + len(self.s) // width + 1):
-            self.scr.move(y, 0)
-            self.scr.clrtoeol()
-
+        self.clear_wrapped_lines()
         self.rl_history.enter(self.s)
         self.s = self.rl_history.forward()
         self.print_line(self.s, clr=True)
@@ -722,7 +766,12 @@ class CLIRepl(Repl):
             self.complete()
             return ''
 
-        elif key == 'KEY_DC':  # Del
+        elif key in key_dispatch[config.delete_key] and not self.s:
+            # Delete on empty line exits
+            self.do_exit = True
+            return None
+
+        elif key in ('KEY_DC', ) + key_dispatch[config.delete_key]:
             self.delete()
             self.complete()
             # Redraw (as there might have been highlighted parens)
@@ -743,12 +792,12 @@ class CLIRepl(Repl):
             self.fwd()
             return ''
 
-        elif key == 'KEY_LEFT':  # Cursor Left
+        elif key in ("KEY_LEFT",' ^B', chr(2)):  # Cursor Left or ^B
             self.mvc(1)
             # Redraw (as there might have been highlighted parens)
             self.print_line(self.s)
 
-        elif key == 'KEY_RIGHT':  # Cursor Right
+        elif key in ("KEY_RIGHT", '^F', chr(6)):  # Cursor Right or ^F
             self.mvc(-1)
             # Redraw (as there might have been highlighted parens)
             self.print_line(self.s)
@@ -763,6 +812,14 @@ class CLIRepl(Repl):
             # Redraw (as there might have been highlighted parens)
             self.print_line(self.s)
 
+        elif key in ("KEY_NPAGE", '\T'): # page_down or \T
+            self.hend()
+            self.print_line(self.s)
+
+        elif key in ("KEY_PPAGE", '\S'): # page_up or \S
+            self.hbegin()
+            self.print_line(self.s)
+
         elif key in key_dispatch[config.cut_to_buffer_key]:  # cut to buffer
             self.cut_to_buffer()
             return ''
@@ -773,7 +830,7 @@ class CLIRepl(Repl):
             return ''
 
         elif key in key_dispatch[config.clear_word_key]:
-            self.bs_word()
+            self.cut_buffer = self.bs_word()
             self.complete()
             return ''
 
@@ -807,19 +864,14 @@ class CLIRepl(Repl):
             return ''
 
         elif key in key_dispatch[config.show_source_key]:
-            try:
-                obj = self.current_func
-                if obj is None and inspection.is_eval_safe_name(self.s):
-                    obj = self.get_object(self.s)
-                source = inspect.getsource(obj)
-            except (AttributeError, IOError, NameError, TypeError):
-                self.statusbar.message("Cannot show source.")
-                return ''
-            else:
+            source = self.get_source_of_current_name()
+            if source is not None:
                 if config.highlight_show_source:
                     source = format(PythonLexer().get_tokens(source),
                                     TerminalFormatter())
                 page(source)
+            else:
+                self.statusbar.message('Cannot show source.')
             return ''
 
         elif key == '\n':
@@ -832,6 +884,10 @@ class CLIRepl(Repl):
         elif key == 'KEY_BTAB':
             return self.tab(back=True)
 
+        elif key in key_dispatch[config.suspend_key]:
+            self.suspend()
+            return ''
+
         elif len(key) == 1 and not unicodedata.category(key) == 'Cc':
             self.addstr(key)
             self.print_line(self.s)
@@ -854,8 +910,7 @@ class CLIRepl(Repl):
             self.highlighted_paren = None
 
         if self.config.syntax and (not self.paste_mode or newline):
-            o = format(self.tokenize(s, newline),
-                       BPythonFormatter(self.config.color_scheme))
+            o = format(self.tokenize(s, newline), self.formatter)
         else:
             o = s
 
@@ -893,9 +948,13 @@ class CLIRepl(Repl):
 
     def push(self, s, insert_into_history=True):
         # curses.raw(True) prevents C-c from causing a SIGINT
-        curses.raw(True)
+        curses.raw(False)
         try:
-            return Repl.push(self, s, insert_into_history)
+            return repl.Repl.push(self, s, insert_into_history)
+        except SystemExit:
+            # Avoid a traceback on e.g. quit()
+            self.do_exit = True
+            return False
         finally:
             curses.raw(True)
 
@@ -989,6 +1048,79 @@ class CLIRepl(Repl):
         self.statusbar.resize(refresh=False)
         self.redraw()
 
+
+    def getstdout(self):
+        """This method returns the 'spoofed' stdout buffer, for writing to a
+        file or sending to a pastebin or whatever."""
+
+        return self.stdout_hist + '\n'
+
+
+    def reevaluate(self):
+        """Clear the buffer, redraw the screen and re-evaluate the history"""
+
+        self.evaluating = True
+        self.stdout_hist = ''
+        self.f_string = ''
+        self.buffer = []
+        self.scr.erase()
+        self.s_hist = []
+        # Set cursor position to -1 to prevent paren matching
+        self.cpos = -1
+
+        self.prompt(False)
+
+        self.iy, self.ix = self.scr.getyx()
+        for line in self.history:
+            if py3:
+                self.stdout_hist += line + '\n'
+            else:
+                self.stdout_hist += line.encode(getpreferredencoding()) + '\n'
+            self.print_line(line)
+            self.s_hist[-1] += self.f_string
+# I decided it was easier to just do this manually
+# than to make the print_line and history stuff more flexible.
+            self.scr.addstr('\n')
+            more = self.push(line)
+            self.prompt(more)
+            self.iy, self.ix = self.scr.getyx()
+
+        self.cpos = 0
+        indent = repl.next_indentation(self.s, self.config.tab_length)
+        self.s = ''
+        self.scr.refresh()
+
+        if self.buffer:
+            for _ in xrange(indent):
+                self.tab()
+
+        self.evaluating = False
+        #map(self.push, self.history)
+        #^-- That's how simple this method was at first :(
+
+    def write(self, s):
+        """For overriding stdout defaults"""
+        if '\x04' in s:
+            for block in s.split('\x04'):
+                self.write(block)
+            return
+        if s.rstrip() and '\x03' in s:
+            t = s.split('\x03')[1]
+        else:
+            t = s
+
+        if not py3 and isinstance(t, unicode):
+            t = t.encode(getpreferredencoding())
+
+        if not self.stdout_hist:
+            self.stdout_hist = t
+        else:
+            self.stdout_hist += t
+
+        self.echo(s)
+        self.s_hist.append(s.rstrip())
+
+
     def show_list(self, items, topline=None, current_item=None):
         shared = Struct()
         shared.cols = 0
@@ -1091,6 +1223,8 @@ class CLIRepl(Repl):
         if v_items:
             self.list_win.addstr('\n ')
 
+        if not py3:
+            encoding = getpreferredencoding()
         for ix, i in enumerate(v_items):
             padding = (wl - len(i)) * ' '
             if i == current_item:
@@ -1098,13 +1232,15 @@ class CLIRepl(Repl):
             else:
                 color = get_colpair(self.config, 'main')
             if not py3:
-                i = i.encode(getpreferredencoding())
+                i = i.encode(encoding)
             self.list_win.addstr(i + padding, color)
             if ((cols == 1 or (ix and not (ix + 1) % cols))
                     and ix + 1 < len(v_items)):
                 self.list_win.addstr('\n ')
 
         if self.docstring is not None:
+            if not py3:
+                docstring_string = docstring_string.encode(encoding, 'ignore')
             self.list_win.addstr('\n' + docstring_string,
                                  get_colpair(self.config, 'comment'))
 # XXX: After all the trouble I had with sizing the list box (I'm not very good
@@ -1137,6 +1273,11 @@ class CLIRepl(Repl):
         self.h = h - 1
         self.x = 0
 
+    def suspend(self):
+        """Suspend the current process for shell job control."""
+        curses.endwin()
+        os.kill(os.getpid(), signal.SIGSTOP)
+
     def tab(self, back=False):
         """Process the tab key being hit. If there's only whitespace
         in the line or the line is blank then process a normal tab,
@@ -1199,15 +1340,15 @@ class CLIRepl(Repl):
         return True
 
     def undo(self, n=1):
-        Repl.undo(self, n)
+        repl.Repl.undo(self, n)
 
         # This will unhighlight highlighted parens
         self.print_line(self.s)
 
     def writetb(self, lines):
         for line in lines:
-            self.echo('\x01%s\x03%s' % (self.config.color_scheme['error'],
-                                        line))
+            self.write('\x01%s\x03%s' % (self.config.color_scheme['error'],
+                                         line))
 
     def yank_from_buffer(self):
         """Paste the text from the cut buffer at the current cursor location"""
@@ -1314,22 +1455,21 @@ class Statusbar(object):
         while True:
             c = self.win.getch()
 
+            # '\b'
             if c == 127:
                 o = bs(o)
-                continue
-
-            if c == 27:
-                raise ValueError
-
-            if not c or c < 0 or c > 127:
-                continue
-            c = chr(c)
-
-            if c == '\n':
+            # '\n'
+            elif c == 10:
                 break
-
-            self.win.addstr(c, get_colpair(self.config, 'prompt'))
-            o += c
+            # ESC
+            elif c == 27:
+                curses.flushinp()
+                raise ValueError
+            # literal
+            elif 0 <= c < 127:
+                c = chr(c)
+                self.win.addstr(c, get_colpair(self.config, 'prompt'))
+                o += c
 
         self.settext(self._s)
         return o
@@ -1380,12 +1520,8 @@ def init_wins(scr, colors, config):
 # Thanks to Angus Gibson for pointing out this missing line which was causing
 # problems that needed dirty hackery to fix. :)
 
-# TODO:
-#
-# This should show to be configured keys from ~/.bpython/config
-#
     statusbar = Statusbar(scr, main_win, background, config,
-        " <%s> Rewind  <%s> Save  <%s> Pastebin  <%s> Pager <%s> Show Source " %
+        " <%s> Rewind  <%s> Save  <%s> Pastebin  <%s> Pager  <%s> Show Source " %
             (config.undo_key, config.save_key,
              config.pastebin_key, config.last_output_key,
              config.show_source_key),
@@ -1398,6 +1534,10 @@ def sigwinch(unused_scr):
     global DO_RESIZE
     DO_RESIZE = True
 
+def sigcont(unused_scr):
+    sigwinch(unused_scr)
+    # Forces the redraw
+    curses.ungetch('')
 
 def gethw():
     """I found this code on a usenet post, and snipped out the bit I needed,
@@ -1515,10 +1655,13 @@ def main_curses(scr, args, config, interactive=True, locals_=None,
     global stdscr
     global DO_RESIZE
     global colors
+    global repl
     DO_RESIZE = False
 
     old_sigwinch_handler = signal.signal(signal.SIGWINCH,
                                          lambda *_: sigwinch(scr))
+    # redraw window after being suspended
+    old_sigcont_handler = signal.signal(signal.SIGCONT, lambda *_: sigcont(scr))
 
     stdscr = scr
     try:
@@ -1539,30 +1682,31 @@ def main_curses(scr, args, config, interactive=True, locals_=None,
     if locals_ is None:
         sys.modules['__main__'] = ModuleType('__main__')
         locals_ = sys.modules['__main__'].__dict__
-    interpreter = Interpreter(locals_, getpreferredencoding())
+    interpreter = repl.Interpreter(locals_, getpreferredencoding())
 
-    repl = CLIRepl(main_win, interpreter, statusbar, config, idle)
-    repl._C = cols
+    clirepl = CLIRepl(main_win, interpreter, statusbar, config, idle)
+    clirepl._C = cols
 
-    sys.stdin = FakeStdin(repl)
-    sys.stdout = repl
-    sys.stderr = repl
-
-    repl.startup()
+    sys.stdin = FakeStdin(clirepl)
+    sys.stdout = clirepl
+    sys.stderr = clirepl
 
     if args:
         bpython.args.exec_code(interpreter, args)
         if not interactive:
             curses.raw(False)
-            return repl.getstdout()
+            return clirepl.getstdout()
+    else:
+        sys.path.insert(0, '')
+        clirepl.startup()
 
     if banner is not None:
-        repl.write(banner)
-        repl.write('\n')
-    repl.repl()
+        clirepl.write(banner)
+        clirepl.write('\n')
+    clirepl.repl()
     if config.hist_length:
         histfilename = os.path.expanduser(config.hist_file)
-        repl.rl_history.save(histfilename, getpreferredencoding())
+        clirepl.rl_history.save(histfilename, getpreferredencoding())
 
     main_win.erase()
     main_win.refresh()
@@ -1570,10 +1714,11 @@ def main_curses(scr, args, config, interactive=True, locals_=None,
     statusbar.win.refresh()
     curses.raw(False)
 
-    # Restore SIGWINCH handler
+    # Restore signal handlers
     signal.signal(signal.SIGWINCH, old_sigwinch_handler)
+    signal.signal(signal.SIGCONT, old_sigcont_handler)
 
-    return repl.getstdout()
+    return clirepl.getstdout()
 
 
 def main(args=None, locals_=None, banner=None):
@@ -1604,6 +1749,7 @@ def main(args=None, locals_=None, banner=None):
 
 
 if __name__ == '__main__':
+    from bpython.cli import main
     main()
 
 # vim: sw=4 ts=4 sts=4 ai et
