@@ -45,6 +45,8 @@ def loadini(struct, configfile):
             'paste_time': 0.02,
             'syntax': True,
             'tab_length': 4,
+            'pastebin_confirm': True,
+            'pastebin_private': False,
             'pastebin_url': 'http://bpaste.net/xmlrpc/',
             'pastebin_show_url': 'http://bpaste.net/show/$paste_id/',
         },
@@ -53,15 +55,20 @@ def loadini(struct, configfile):
             'clear_screen': 'C-l',
             'clear_word': 'C-w',
             'cut_to_buffer': 'C-k',
+            'delete': 'C-d',
             'down_one_line': 'C-n',
-            'exit': 'C-d',
+            'exit': '',
             'last_output': 'F9',
             'pastebin': 'F8',
             'save': 'C-s',
             'show_source': 'F2',
+            'suspend': 'C-z',
             'undo': 'C-r',
             'up_one_line': 'C-p',
-            'yank_from_buffer': 'C-y'}})
+            'yank_from_buffer': 'C-y'},
+        'gtk': {
+            'font': 'monospace 10',
+            'color_scheme': 'default'}})
     config.read(config_path)
 
     struct.dedent_after = config.getint('general', 'dedent_after')
@@ -79,6 +86,7 @@ def loadini(struct, configfile):
     struct.pastebin_key = config.get('keyboard', 'pastebin')
     struct.save_key = config.get('keyboard', 'save')
     struct.show_source_key = config.get('keyboard', 'show_source')
+    struct.suspend_key = config.get('keyboard', 'suspend')
     struct.undo_key = config.get('keyboard', 'undo')
     struct.up_one_line_key = config.get('keyboard', 'up_one_line')
     struct.down_one_line_key = config.get('keyboard', 'down_one_line')
@@ -87,13 +95,19 @@ def loadini(struct, configfile):
     struct.clear_word_key = config.get('keyboard', 'clear_word')
     struct.clear_line_key = config.get('keyboard', 'clear_line')
     struct.clear_screen_key = config.get('keyboard', 'clear_screen')
+    struct.delete_key = config.get('keyboard', 'delete')
     struct.exit_key = config.get('keyboard', 'exit')
     struct.last_output_key = config.get('keyboard', 'last_output')
 
+    struct.pastebin_confirm = config.getboolean('general', 'pastebin_confirm')
+    struct.pastebin_private = config.getboolean('general', 'pastebin_private')
     struct.pastebin_url = config.get('general', 'pastebin_url')
     struct.pastebin_show_url = config.get('general', 'pastebin_show_url')
 
+    struct.gtk_font = config.get('gtk', 'font')
+
     color_scheme_name = config.get('general', 'color_scheme')
+    color_gtk_scheme_name = config.get('gtk', 'color_scheme')
 
     default_colors = {
             'keyword': 'y',
@@ -113,17 +127,43 @@ def loadini(struct, configfile):
             'prompt_more': 'g',
         }
  
+    default_gtk_colors = {
+            'keyword': 'b',
+            'name': 'k',
+            'comment': 'b',
+            'string': 'm',
+            'error': 'r',
+            'number': 'G',
+            'operator': 'B',
+            'punctuation': 'g',
+            'token': 'C',
+            'background': 'w',
+            'output': 'k',
+            'main': 'c',
+            'paren': 'R',
+            'prompt': 'b',
+            'prompt_more': 'g',
+        }
+
+    # TODO consolidate
     if color_scheme_name == 'default':
         struct.color_scheme = default_colors
     else:
         path = os.path.expanduser('~/.bpython/%s.theme' % (color_scheme_name,))
         load_theme(struct, path, config_path, default_colors)
 
+    if color_gtk_scheme_name == 'default':
+        struct.color_gtk_scheme = default_gtk_colors
+    else:
+        path = os.path.expanduser('~/.bpython/%s.theme' % (color_gtk_scheme_name,))
+        load_gtk_theme(struct, path, config_path, default_gtk_colors)
+
+
     # checks for valid key configuration this part still sucks
     for key in (struct.pastebin_key, struct.save_key):
         key_dispatch[key]
 
-
+# TODO consolidate
 def load_theme(struct, path, inipath, default_colors):
     theme = ConfigParser()
     try:
@@ -147,6 +187,30 @@ def load_theme(struct, path, inipath, default_colors):
     f.close()
 
 
+def load_gtk_theme(struct, path, inipath, default_colors):
+    theme = ConfigParser()
+    try:
+        f = open(path, 'r')
+    except (IOError, OSError), e:
+        sys.stdout.write("Error loading theme file specified in '%s':\n%s\n" %
+                         (inipath, e))
+        sys.exit(1)
+    theme.readfp(f)
+    struct.color_gtk_scheme = {}
+    for k, v in chain(theme.items('syntax'), theme.items('interface')):
+        if theme.has_option('syntax', k):
+            struct.color_gtk_scheme[k] = theme.get('syntax', k)
+        else:
+            struct.color_gtk_scheme[k] = theme.get('interface', k)
+
+    # Check against default theme to see if all values are defined
+    for k, v in default_colors.iteritems():
+        if k not in struct.color_gtk_scheme:
+            struct.color_gtk_scheme[k] = v
+    f.close()
+
+
+
 def migrate_rc(path):
     """Use the shlex module to convert the old configuration file to the new
     format.
