@@ -1,6 +1,6 @@
 # The MIT License
 #
-# Copyright (c) 2009 the bpython authors.
+# Copyright (c) 2009-2010 the bpython authors.
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -22,7 +22,9 @@
 #
 
 from __future__ import with_statement
+import collections
 import inspect
+import keyword
 import pydoc
 import re
 import sys
@@ -31,11 +33,21 @@ import types
 from pygments.lexers import PythonLexer
 from pygments.token import Token
 
+try:
+    collections.Callable
+    has_collections_callable = True
+except AttributeError:
+    has_collections_callable = False
+try:
+    types.InstanceType
+    has_instance_type = True
+except AttributeError:
+    has_instance_type = False
 
 py3 = sys.version_info[0] == 3
 
 if not py3:
-    _name = re.compile(r'[a-zA-Z_]\w*')
+    _name = re.compile(r'[a-zA-Z_]\w*$')
 
 
 class AttrCleaner(object):
@@ -197,7 +209,14 @@ def getpydocspec(f, func):
 
 
 def getargspec(func, f):
-    is_bound_method = inspect.ismethod(f) and f.im_self is not None
+    # Check if it's a real bound method or if it's implicitly calling __init__
+    # (i.e. FooClass(...) and not FooClass.__init__(...) -- the former would
+    # not take 'self', the latter would:
+    func_name = getattr(f, '__name__', None)
+
+    is_bound_method = ((inspect.ismethod(f) and f.im_self is not None)
+                    or (func_name == '__init__' and not
+                        func.endswith('.__init__')))
     try:
         if py3:
             argspec = inspect.getfullargspec(f)
@@ -220,6 +239,18 @@ def getargspec(func, f):
 
 def is_eval_safe_name(string):
     if py3:
-        return all(part.isidentifier() for part in string.split('.'))
+        return all(part.isidentifier() and not keyword.iskeyword(part)
+                   for part in string.split('.'))
     else:
-        return all(_name.match(part) for part in string.split('.'))
+        return all(_name.match(part) and not keyword.iskeyword(part)
+                   for part in string.split('.'))
+
+
+def is_callable(obj):
+    if has_instance_type and isinstance(obj, types.InstanceType):
+        # Work around a Python bug, see issue 7624
+        return callable(obj)
+    elif has_collections_callable:
+        return isinstance(obj, collections.Callable)
+    else:
+        return callable(obj)
