@@ -1,6 +1,6 @@
 # The MIT License
 #
-# Copyright (c) 2009 the bpython authors.
+# Copyright (c) 2009-2010 the bpython authors.
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -21,7 +21,6 @@
 # THE SOFTWARE.
 #
 
-# G*RLLCRA RHTNSAHTNSOG()*(@****
 from __future__ import with_statement
 import code
 import codecs
@@ -36,6 +35,7 @@ import traceback
 from glob import glob
 from itertools import takewhile
 from locale import getpreferredencoding
+from socket import error as SocketError
 from string import Template
 from urllib import quote as urlquote
 from xmlrpclib import ServerProxy, Error as XMLRPCError
@@ -46,12 +46,22 @@ from pygments.token import Token
 from bpython import importcompletion, inspection
 from bpython.formatter import Parenthesis
 
+# Needed for special handling of __abstractmethods__
+# abc only exists since 2.6, so check both that it exists and that it's
+# the one we're expecting
+try:
+    import abc
+    abc.ABCMeta
+    has_abc = True
+except (ImportError, AttributeError):
+    has_abc = False
+
 py3 = sys.version_info[0] == 3
 
 
 class Interpreter(code.InteractiveInterpreter):
 
-    def __init__(self, locals=None, encoding=sys.getdefaultencoding()):
+    def __init__(self, locals=None, encoding=None):
         """The syntaxerror callback can be set at any time and will be called
         on a caught syntax error. The purpose for this in bpython is so that
         the repl can be instantiated after the interpreter (which it
@@ -60,7 +70,7 @@ class Interpreter(code.InteractiveInterpreter):
         specifically, this is so that autoindentation does not occur after a
         traceback."""
 
-        self.encoding = encoding
+        self.encoding = encoding or sys.getdefaultencoding()
         self.syntaxerror_callback = None
 # Unfortunately code.InteractiveInterpreter is a classic class, so no super()
         code.InteractiveInterpreter.__init__(self, locals)
@@ -149,22 +159,20 @@ class History(object):
         if line:
             self.entries.append(line)
 
+    def first(self):
+        """Move back to the beginning of the history."""
+        if not self.is_at_end:
+            self.index = len(self.entries)
+        return self.entries[-self.index]
+
     def back(self):
-        if not self.is_at_end():
+        """Move one step back in the history."""
+        if not self.is_at_end:
             self.index += 1
         return self.entries[-self.index]
 
-    def is_at_end(self):
-        return self.index >= len(self.entries)
-
-    def is_at_start(self):
-        return self.index == 0
-
-    def enter(self, line):
-        if self.index == 0:
-            self.saved_line = line
-
     def forward(self):
+        """Move one step forward in the history."""
         if self.index > 1:
             self.index -= 1
             return self.entries[-self.index]
@@ -172,6 +180,24 @@ class History(object):
             self.index = 0
             return self.saved_line
 
+    def last(self):
+        """Move forward to the end of the history."""
+        if not self.is_at_start:
+            self.index = 0
+        return self.entries[0]
+
+    @property
+    def is_at_end(self):
+        return self.index >= len(self.entries) or self.index == -1
+
+    @property
+    def is_at_start(self):
+        return self.index == 0
+
+    def enter(self, line):
+        if self.index == 0:
+            self.saved_line = line
+
     @classmethod
     def from_filename(cls, filename):
         history = cls()
@@ -217,7 +243,10 @@ class MatchesIterator(object):
         return self.matches[self.index]
 
     def previous(self):
-        self.index = (self.index - 1) % len(self.matches)
+        if self.index <= 0:
+            self.index = len(self.matches)
+        self.index -= 1
+
         return self.matches[self.index]
 
     def update(self, current_word='', matches=[]):
@@ -227,6 +256,23 @@ class MatchesIterator(object):
             self.index = -1
 
 
+class Interaction(object):
+    def __init__(self, config, statusbar=None):
+        self.config = config
+
+        if statusbar:
+            self.statusbar = statusbar
+
+    def confirm(self, s):
+        raise NotImplementedError
+
+    def notify(self, s, n=10):
+        raise NotImplementedError
+
+    def file_prompt(self, s):
+        raise NotImplementedError
+
+
 class Repl(object):
     """Implements the necessary guff for a Python-repl-alike interface
 
@@ -260,15 +306,13 @@ class Repl(object):
     XXX Subclasses should implement echo, current_line, cw
     """
 
-    def __init__(self, interp, config, idle=None):
-        """Initialise the repl with, unfortunately, a curses screen passed to
-        it.  This needs to be split up so the curses crap isn't in here.
+    def __init__(self, interp, config):
+        """Initialise the repl.
 
         interp is a Python code.InteractiveInterpreter instance
 
-        The optional 'idle' parameter is a function that the repl call while
-        it's blocking (waiting for keypresses). This, again, should be in a
-        different class"""
+        config is a populated bpython.config.Struct.
+        """
 
         self.config = config
         self.cut_buffer = ''
@@ -277,7 +321,6 @@ class Repl(object):
         self.interp.syntaxerror_callback = self.clear_current_line
         self.match = False
         self.rl_history = History()
-        self.stdout_hist = ''
         self.s_hist = []
         self.history = []
         self.evaluating = False
@@ -296,7 +339,14 @@ class Repl(object):
         self.list_win_visible = False
         self._C = {}
         self.prev_block_finished = 0
-        sys.path.insert(0, '.')
+        self.interact = Interaction(self.config)
+        # previous pastebin content to prevent duplicate pastes, filled on call
+        # to repl.pastebin
+        self.prev_pastebin_content = ''
+        self.prev_pastebin_url = ''
+        # Necessary to fix mercurial.ui.ui expecting sys.stderr to have this
+        # attribute
+        self.closed = False
 
         pythonhist = os.path.expanduser(self.config.hist_file)
         if os.path.exists(pythonhist):
@@ -340,6 +390,11 @@ class Repl(object):
         if hasattr(obj, '__class__'):
             words.append('__class__')
             words = words + rlcompleter.get_class_members(obj.__class__)
+            if has_abc and not isinstance(obj.__class__, abc.ABCMeta):
+                try:
+                    words.remove('__abstractmethods__')
+                except ValueError:
+                    pass
 
         matches = []
         n = len(attr)
@@ -351,7 +406,7 @@ class Repl(object):
     def _callable_postfix(self, value, word):
         """rlcompleter's _callable_postfix done right."""
         with inspection.AttrCleaner(value):
-            if hasattr(value, '__call__'):
+            if inspection.is_callable(value):
                 word += '('
         return word
 
@@ -453,6 +508,22 @@ class Repl(object):
             return True
         return False
 
+    def get_source_of_current_name(self):
+        """Return the source code of the object which is bound to the
+        current name in the current input line. Return `None` if the
+        source cannot be found."""
+        try:
+            obj = self.current_func
+            if obj is None:
+                line = self.current_line()
+                if inspection.is_eval_safe_name(line):
+                    obj = self.get_object(line)
+            source = inspect.getsource(obj)
+        except (AttributeError, IOError, NameError, TypeError):
+            return None
+        else:
+            return source
+
     def complete(self, tab=False):
         """Construct a full list of possible completions and construct and
         display them in a window. Also check if there's an available argspec
@@ -580,12 +651,6 @@ class Repl(object):
             indentation = 0
         return indentation
 
-    def getstdout(self):
-        """This method returns the 'spoofed' stdout buffer, for writing to a
-        file or sending to a pastebin or whatever."""
-
-        return self.stdout_hist + '\n'
-
     def formatforfile(self, s):
         """Format the stdout buffer to something suitable for writing to disk,
         i.e. without >>> and ... at input lines and with "# OUT: " prepended to
@@ -604,9 +669,12 @@ class Repl(object):
         buffer to disk."""
 
         try:
-            fn = self.statusbar.prompt('Save to file (Esc to cancel): ')
+            fn = self.interact.file_prompt('Save to file (Esc to cancel): ')
+            if not fn:
+                self.interact.notify("Save cancelled.")
+                return
         except ValueError:
-            self.statusbar.message("Save cancelled.")
+            self.interact.notify("Save cancelled.")
             return
 
         if fn.startswith('~'):
@@ -619,28 +687,43 @@ class Repl(object):
             f.write(s)
             f.close()
         except IOError:
-            self.statusbar.message("Disk write error for file '%s'." % (fn, ))
+            self.interact.notify("Disk write error for file '%s'." % (fn, ))
         else:
-            self.statusbar.message('Saved to %s' % (fn, ))
+            self.interact.notify('Saved to %s' % (fn, ))
 
-    def pastebin(self):
+    def pastebin(self, s=None):
         """Upload to a pastebin and display the URL in the status bar."""
 
+        if s is None:
+            s = self.getstdout()
+
+        if (self.config.pastebin_confirm and
+            not self.interact.confirm("Pastebin buffer? (y/N) ")):
+            self.interact.notify("Pastebin aborted")
+            return
+
         pasteservice = ServerProxy(self.config.pastebin_url)
 
-        s = self.getstdout()
+        if s == self.prev_pastebin_content:
+            self.interact.notify('Duplicate pastebin. Previous URL: ' +
+                                  self.prev_pastebin_url)
+            return self.prev_pastebin_url
 
-        self.statusbar.message('Posting data to pastebin...')
+        self.prev_pastebin_content = s
+
+        self.interact.notify('Posting data to pastebin...')
         try:
             paste_id = pasteservice.pastes.newPaste('pycon', s)
-        except XMLRPCError, e:
-            self.statusbar.message('Upload failed: %s' % (str(e), ) )
+        except (SocketError, XMLRPCError), e:
+            self.interact.notify('Upload failed: %s' % (str(e), ) )
             return
 
         paste_url_template = Template(self.config.pastebin_show_url)
         paste_id = urlquote(paste_id)
         paste_url = paste_url_template.safe_substitute(paste_id=paste_id)
-        self.statusbar.message('Pastebin URL: %s' % (paste_url, ), 10)
+        self.prev_pastebin_url = paste_url
+        self.interact.notify('Pastebin URL: %s' % (paste_url, ), 10)
+        return paste_url
 
     def push(self, s, insert_into_history=True):
         """Push a line of code onto the buffer so it can process it all
@@ -651,12 +734,7 @@ class Repl(object):
         if insert_into_history:
             self.rl_history.append(s)
 
-        try:
-            more = self.interp.runsource('\n'.join(self.buffer))
-        except SystemExit:
-            # Avoid a traceback on e.g. quit()
-            self.do_exit = True
-            return False
+        more = self.interp.runsource('\n'.join(self.buffer))
 
         if not more:
             self.buffer = []
@@ -682,70 +760,6 @@ class Repl(object):
 
         self.rl_history.entries = entries
 
-    def reevaluate(self):
-        """Clear the buffer, redraw the screen and re-evaluate the history"""
-
-        self.evaluating = True
-        self.stdout_hist = ''
-        self.f_string = ''
-        self.buffer = []
-        self.scr.erase()
-        self.s_hist = []
-        # Set cursor position to -1 to prevent paren matching
-        self.cpos = -1
-
-        self.prompt(False)
-
-        self.iy, self.ix = self.scr.getyx()
-        for line in self.history:
-            if py3:
-                self.stdout_hist += line + '\n'
-            else:
-                self.stdout_hist += line.encode(getpreferredencoding()) + '\n'
-            self.print_line(line)
-            self.s_hist[-1] += self.f_string
-# I decided it was easier to just do this manually
-# than to make the print_line and history stuff more flexible.
-            self.scr.addstr('\n')
-            more = self.push(line)
-            self.prompt(more)
-            self.iy, self.ix = self.scr.getyx()
-
-        self.cpos = 0
-        indent = next_indentation(self.s, self.config.tab_length)
-        self.s = ''
-        self.scr.refresh()
-
-        if self.buffer:
-            for _ in xrange(indent):
-                self.tab()
-
-        self.evaluating = False
-        #map(self.push, self.history)
-        #^-- That's how simple this method was at first :(
-
-    def write(self, s):
-        """For overriding stdout defaults"""
-        if '\x04' in s:
-            for block in s.split('\x04'):
-                self.write(block)
-            return
-        if s.rstrip() and '\x03' in s:
-            t = s.split('\x03')[1]
-        else:
-            t = s
-
-        if not py3 and isinstance(t, unicode):
-            t = t.encode(getpreferredencoding())
-
-        if not self.stdout_hist:
-            self.stdout_hist = t
-        else:
-            self.stdout_hist += t
-
-        self.echo(s)
-        self.s_hist.append(s.rstrip())
-
     def flush(self):
         """Olivier Grisel brought it to my attention that the logging
         module tries to call this method, since it makes assumptions
